@@ -1,4 +1,4 @@
-import imp
+import datetime
 import os
 import re
 import sys
@@ -6,11 +6,12 @@ from fabric.api import (hide, local, env, execute, abort, lcd, sudo, run, cd, pu
 from fabric.colors import red, green, yellow, cyan
 from fabric.contrib import django as fab_django
 from fabric.contrib.console import confirm
-from fabric.contrib.files import upload_template
+from fabric.contrib.files import exists as remote_exists
 from fabric.decorators import task, roles
 from fabric.state import output as output_level
 from fabric.utils import error
-from eggplant.utils.fab import lsudo, virtual_env
+from eggplant.utils import SettingsDict
+from eggplant.utils.fab import lsudo, virtual_env, fab_argument
 from eggplant.utils.filesystem import temporary_directory, rsync, render_file
 
 # Set env
@@ -21,7 +22,7 @@ env.roledefs.update({
 
 # Private Functions
 
-def deploy_setting(project_name=None, deploy_settings_path=None):
+def deploy_setting(project_name=None, deploy_settings_path=None, **setting_patches):
     """Get deploy_settings.
     This method will be called by deploy task.
     If you defined one of project_name and deploy_settings in fabric's env, you don't have to call this.
@@ -51,7 +52,13 @@ def deploy_setting(project_name=None, deploy_settings_path=None):
         print error(red('Cannot find deploy settings', bold=True))
 
     # Load deploy_settings
-    env.deploy_settings = imp.load_source('', deploy_settings_path)
+    default_deploy_settings_path = os.path.join(os.path.dirname(__file__), 'default_deploy_settings.py')
+    if project_name is not None:
+        setting_patches.update({'PROJECT_NAME': project_name})
+    patches = {}
+    for key, value in setting_patches.items():
+        patches[key] = fab_argument(value)
+    env.deploy_settings = SettingsDict(default_deploy_settings_path, deploy_settings_path, **patches)
 
     # Find project name - 2
     if project_name is None:
@@ -64,8 +71,9 @@ def deploy_setting(project_name=None, deploy_settings_path=None):
 
 # Tasks
 
-@task
-def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', verbose=0):
+@task(default=True)
+def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy',
+           verbose=0, update_apache_conf=False, **additional_settings):
     """Deploy current django project to targeted web servers
 
     This is a public task. It will prepare arguments for the
@@ -86,14 +94,19 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
         The message used for git commit in deploy target
     :param verbose:
         Print information or not. Default is 0. Levels=[0, 1, 2, 3]
+    :param update_apache_conf:
+        Update Apache configuration file or not.
+        If this is first-time deploy, the value will be forced to True.
     """
     # Get deploy settings and project name
-    execute(deploy_setting, project_name=project_name, deploy_settings_path=deploy_settings_path)
+    execute(deploy_setting, project_name=project_name,
+        deploy_settings_path=deploy_settings_path, **additional_settings)
     project_name = env.project_name
     deploy_settings = env.deploy_settings
+    env.update_apache_conf = update_apache_conf
 
     # Check output level
-    verbose = int(verbose)
+    env.verbose = verbose = int(verbose)
     if verbose==0:
         hide_output_status = ['everything']
     elif verbose==1:
@@ -106,22 +119,24 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
         setattr(output_level, status, False)
 
     # Load ssh config
-    env.use_ssh_config = getattr(deploy_settings, 'USE_SSH_CONFIG', False)
+    env.use_ssh_config = deploy_settings.get('USE_SSH_CONFIG')
 
     # Load web servers
     if len(env.roledefs['web_servers'])==0:
         with hide('everything'):
-            execute('web_server', *getattr(deploy_settings, 'WEB_SERVERS', []))
+            execute('web_server', *deploy_settings.get('WEB_SERVERS'))
             if verbose!=0: puts('')
+    if len(env.roledefs['web_servers'])==0:
+        abort('Where do you want to deploy to?')
 
     with temporary_directory(prefix='deploy~%s~'%project_name) as tmp_directory:
         env.tmp_directory = tmp_directory
         # Copy codebase
         puts(cyan('Copy your codebase to temp directory'))
-        original_src_root = getattr(deploy_settings, 'SOURCE_ROOT', os.getcwd())
+        original_src_root = deploy_settings.get('SOURCE_ROOT', os.getcwd())
         # Rsync
         copy_codebase_rsync_args = '-a --delete --exclude .git'
-        ignore_file_list = getattr(deploy_settings, 'IGNORE_FILE_LIST', os.path.join(original_src_root, '.gitignore'))
+        ignore_file_list = deploy_settings.get('IGNORE_FILE_LIST', os.path.join(original_src_root, '.gitignore'))
         if ignore_file_list is not None and os.path.exists(ignore_file_list):
             copy_codebase_rsync_args += ' --exclude-from \'%s\''%os.path.relpath(ignore_file_list, os.getcwd())
         if not original_src_root.endswith('/'): original_src_root += '/'
@@ -138,18 +153,16 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
 
         with lcd(tmp_directory):
             # Switch settings
-            project_settings_module = getattr(deploy_settings, 'PROJECT_SETTINGS', '%s.settings'%project_name)
-            project_settings_path = project_settings_module.replace('.', '/')+'.py'
+            project_settings_module = deploy_settings.get('PROJECT_SETTINGS', '%s.settings'%project_name)
+            env.project_settings_path = project_settings_path = project_settings_module.replace('.', '/')+'.py'
             # Use symbolic link or not?
-            use_symbolic_link_settings = getattr(
-                deploy_settings, 'USE_SYMBOLIC_LINK_SETTINGS', os.path.islink(project_settings_path)
-            )
+            use_symbolic_link_settings = deploy_settings.get('USE_SYMBOLIC_LINK_SETTINGS',
+                os.path.islink(project_settings_path))
             if use_symbolic_link_settings:
                 # Relink to production one
                 puts(cyan('Move to production settings'))
-                production_settings_module = getattr(
-                    deploy_settings, 'PRODUCTION_SETTINGS', '%s.production_settings'%project_name
-                )
+                production_settings_module = deploy_settings.get('PRODUCTION_SETTINGS',
+                    '%s.production_settings'%project_name)
                 production_settings_path = production_settings_module.replace('.', '/')+'.py'
                 local('rm -rf %s'%project_settings_path)
                 local('ln -s %s %s'%(
@@ -165,7 +178,7 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
             env.django_settings = django_settings
 
             # Set paths
-            remote_base_path = getattr(deploy_settings, 'REMOTE_BASE', '/var/www/django')
+            remote_base_path = deploy_settings.get('REMOTE_BASE')
             env.remote_base_path = remote_base_path
 
             # Get Project Information
@@ -183,11 +196,11 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
             print ''
 
             # Check TODOs and FIX_MEs
-            if getattr(deploy_settings, 'CHECK_COMMENTS', True):
+            if deploy_settings.get('CHECK_COMMENTS'):
                 print cyan('Check comment in source code')
-                check_todo = getattr(deploy_settings, 'CHECK_TODO', False)
-                check_note = getattr(deploy_settings, 'CHECK_NOTE', False)
-                check_fix_me = getattr(deploy_settings, 'CHECK_FIX_ME', True)
+                check_todo = deploy_settings.get('CHECK_TODO')
+                check_note = deploy_settings.get('CHECK_NOTE')
+                check_fix_me = deploy_settings.get('CHECK_FIX_ME')
                 comment_types = []
                 if check_todo: comment_types += ['TODO']
                 if check_note: comment_types += ['NOTE']
@@ -225,7 +238,7 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
                 print ''
 
             # Ask
-            if getattr(deploy_settings, 'CONFIRM_BEFORE_DEPLOY', True):
+            if deploy_settings.get('CONFIRM_BEFORE_DEPLOY'):
                 try:
                     if not confirm('deploy?', default=False):
                         raise KeyboardInterrupt
@@ -242,7 +255,7 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
             for gitignore_item in raw_gitignores:
                 if pyc_pattern.match(gitignore_item) is None:
                     gitignores += [gitignore_item]
-            gitignores += getattr(deploy_settings, 'ADDITIONAL_IGNORE_LIST', [])
+            gitignores += deploy_settings.get('ADDITIONAL_IGNORE_LIST')
             gitignores += [
                 'logs/',
             ]
@@ -252,7 +265,7 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
                 f.write(processed_gitignore_content)
 
             # Compile static
-            if getattr(deploy_settings, 'COLLECT_STATIC', True):
+            if deploy_settings.get('COLLECT_STATIC'):
                 puts(cyan('Compile static files'))
                 with lcd(project_src_root):
                     if 'collectstatic' in get_commands():
@@ -262,79 +275,10 @@ def deploy(project_name=None, deploy_settings_path=None, commit_msg='deploy', ve
                         local('python manage.py compress')
                 if verbose!=0: puts('')
 
-            # Make WSGI
-            if getattr(deploy_settings, 'USE_MOD_WSGI', True):
-                puts(cyan('Generate configuration file of mod_wsgi and Apache'))
-                # Get Apache conf template
-                apache_conf_template_filename = 'apache.conf.jinja2'
-                apache_conf_template_path = getattr(deploy_settings, 'APACHE_CONF_TEMPLATE', None)
-                if apache_conf_template_path is None:
-                    apache_conf_template_path = os.path.join(
-                        project_src_root,
-                        os.path.dirname(project_settings_path),
-                        apache_conf_template_filename
-                    )
-                if not os.path.exists(apache_conf_template_path):
-                    apache_conf_template_path = os.path.join(
-                        os.path.dirname(__file__),
-                        'deploy/%s'%apache_conf_template_filename
-                    )
-                # Make context for apache conf
-                source_root_name = env.source_root_name = 'source'
-                venv_name = env.venv_name = 'venv'
-                remote_project_path = env.remote_project_path = os.path.join(remote_base_path, project_name)
-                remote_project_source_root = env.remote_project_source_root = \
-                    os.path.join(remote_project_path, source_root_name)
-                remote_project_venv = env.remote_project_venv = os.path.join(remote_project_path, venv_name)
-                python_version = sys.version_info
-                default_site_packages = os.path.join(
-                    remote_project_venv,
-                    'lib/python%s.%s/site-packages'%(python_version[0], python_version[1])
-                )
-                context_keys = [
-                    ('AWS_S3_BASE_URL', None),
-                    ('MEDIA_URL', '/media/'),
-                    ('MEDIA_PATH', os.path.join(remote_project_source_root, 'media')),
-                    ('REDIRECT_STATIC_TO_S3', False),
-                    ('STATIC_URL', '/static/'),
-                    ('STATIC_PATH', os.path.join(remote_project_source_root, 'static')),
-                    ('SERVE_FAVICON', False),
-                    ('SERVE_MEDIA', False),
-                    ('SERVE_ROBOTS', False),
-                    ('SERVE_STATIC', True),
-                    ('URL_PREFIX', '/'),
-                    ('VIRTUALENV_SITE_PACKAGES_PATH', default_site_packages),
-                    ('WSGI_SCRIPT_DIR_NAME', os.path.join(remote_project_source_root, project_name)),
-                    ('WSGI_SCRIPT_FILE_NAME', 'wsgi.py'),
-                    ('WSGI_PROCESS_GROUP', 'apache'),
-                    ('WSGI_PROCESS_NAME', project_name),
-                    ('WSGI_PROCESS_USER', 'apache'),
-                    ('WSGI_PROCESSES_COUNT', 2),
-                    ('WSGI_PYTHON_OPTIMIZE_LEVEL', 1),
-                    ('WSGI_THREADS_COUNT', 15),
-                ]
-                apache_conf_context = {}
-                for context_key, context_default_value in context_keys:
-                    context_value = getattr(deploy_settings, context_key, context_default_value)
-                    apache_conf_context[context_key.lower()] = context_value
-                apache_conf_context['project_path'] = remote_project_source_root
-                # Render Apache conf
-                apache_conf_content = render_file(apache_conf_template_path, apache_conf_context).strip()
-                apache_conf_content = re.sub(r'( *\n){2,}', '\n\n', apache_conf_content) + '\n'
-                # Write it temporarily
-                apache_conf_temp_path = os.path.join(
-                    project_src_root,
-                    os.path.dirname(project_settings_path),
-                    'apache.conf'
-                )
-                env.tmp_apache_conf_path = os.path.relpath(apache_conf_temp_path, project_src_root)
-                with open(apache_conf_temp_path, 'w') as f:
-                    f.write(apache_conf_content)
-
             # Go!
             print cyan('Send codebase to remote servers')
             print 'Remote: %s'%env.roledefs['web_servers']
-            if verbose!=0: puts('')
+            print ''
             # Save arguments
             env.commit_msg = commit_msg
             execute(deploy_to_remote)
@@ -348,106 +292,203 @@ def deploy_to_remote():
     This is the private core task who really do deploy to web servers.
     Call "deploy" task. Don't call this directly.
     """
+    # local inner function
+    def print_title(title):
+        if verbose==0: print title
+        else: print cyan(title)
+
     # Get variables
+    remote_base_path = env.remote_base_path
     deploy_settings = env.deploy_settings
     project_name = env.project_name
+    source_root_name = 'source'
+    venv_name = 'venv'
+    remote_project_path = os.path.join(remote_base_path, project_name)
+    remote_project_source_root = env.remote_project_source_root =\
+    os.path.join(remote_project_path, source_root_name)
+    remote_project_venv = env.remote_project_venv = os.path.join(remote_project_path, venv_name)
     project_src_root = env.project_src_root
-    remote_base_path = env.remote_base_path
     host_string = env.host_string
     commit_msg = env.commit_msg
-
-    # Define function callee
+    verbose = env.verbose
     host_name = host_string.split('@')[-1]
-    use_sudo = getattr(deploy_settings, 'USE_SUDO_IN_REMOTE', True)
+    project_settings_path = env.project_settings_path
+    update_apache_conf = env.update_apache_conf
+
+    # Define function to call
     if host_name in ('127.0.0.1', 'localhost'):
-        # run
-        if use_sudo: target_run = lsudo
-        else: target_run = local
+        is_remote = False
+        capture_output = {'capture': True}
+        target_run = local
         # sudo
         target_sudo = lsudo
         # cd
         target_cd = lcd
+        # file existence
+        target_file_existence = os.path.exists
     else:
-        # run
-        if use_sudo: target_run = sudo
-        else: target_run = run
+        is_remote = True
+        capture_output = {}
+        target_run = run
         # sudo
         target_sudo = sudo
         # cd
         target_cd = cd
+        # file existence
+        target_file_existence = remote_exists
 
-    puts(green('Deploy to %s'%host_string))
+    # Get env
+    with hide('everything'):
+        remote_user = target_run('whoami', **capture_output)
+        remote_group = target_run('id -g -n %s'%remote_user, **capture_output)
+
+    print(yellow('Deploy to %s'%host_string, bold=True))
 
     # Make room for deploy target
     puts('Deploy target is: %s'%remote_base_path)
-    if not os.path.exists(remote_base_path):
-        puts(yellow('Deploy target doesn\'t exist. Create one now. (The target will be cleaned if existed)'))
-        target_run('mkdir -p %s'%remote_base_path)
-        target_run('chmod o+rx %s'%remote_base_path)
-
-    # Make room for your project
-    source_root_name = env.source_root_name
-    remote_project_path = env.remote_project_path
-    remote_project_source_root = env.remote_project_source_root
-    remote_project_venv = env.remote_project_venv
+    if not target_file_existence(remote_base_path):
+        print(yellow('Deploy target doesn\'t exist. Create one now.'))
+        target_sudo('mkdir -p %s'%remote_base_path)
+        target_sudo('chmod o+rx %s'%remote_base_path)
 
     # Init project
-    if not os.path.exists(remote_project_path):
-        puts(yellow('First time deploy, create git repository:'))
+    if not target_file_existence(remote_project_path):
+        print(yellow('First time deploy, create git repository:'))
+        update_apache_conf = True
         # Make room
-        target_run('mkdir -p %s'%remote_project_path)
+        target_sudo('mkdir -p %s'%remote_project_path)
+        target_sudo('chown -R %s:%s %s'%(remote_user, remote_group, remote_project_path))
         # Source folder with git
         target_run('mkdir -p %s'%remote_project_source_root)
         with target_cd(remote_project_path):
             target_run('git init')
-            target_run('touch .gitignore')
+            target_run('echo "# ignore files" >> .gitignore')
             target_run('git add .gitignore && git commit -m "Init commit"')
         # Virtualenv
         puts(yellow('First time deploy, initialize virtualenv:'))
         with target_cd(remote_project_path):
-            target_run('virtualenv venv')
+            target_run('virtualenv %s'%venv_name)
+
+    # Install apache conf
+    wsgi_script_dir_name = deploy_settings.get('WSGI_SCRIPT_DIR_NAME',
+        os.path.join(remote_project_source_root, project_name))
+    wsgi_script_file_name = deploy_settings.get('WSGI_SCRIPT_FILE_NAME')
+    apache_conf_tmp_name = 'apache~%s.conf'%host_name
+    if deploy_settings.get('INSTALL_APACHE_CONF') and update_apache_conf:
+        puts(cyan('Generate configuration file of mod_wsgi and Apache'))
+        # Get Apache conf template
+        apache_conf_template_filename = 'apache.conf.jinja2'
+        apache_conf_template_path = deploy_settings.get('APACHE_CONF_TEMPLATE')
+        if apache_conf_template_path is None:
+            apache_conf_template_path = os.path.join(
+                project_src_root,
+                os.path.dirname(project_settings_path),
+                apache_conf_template_filename
+            )
+        if not os.path.exists(apache_conf_template_path):
+            apache_conf_template_path = os.path.join(
+                os.path.dirname(__file__),
+                'deploy/%s' % apache_conf_template_filename
+            )
+        # Make context for apache conf
+        with hide('everything'):
+            raw_python_version = target_run('python -c "import sys; print sys.version_info"', **capture_output)
+            python_version = map(lambda x:x.strip().replace('\'', ''), raw_python_version[1:-1].split(','))
+        default_site_packages = os.path.join(
+            remote_project_venv,
+            'lib/python%s.%s/site-packages' % (python_version[0], python_version[1])
+        )
+        apache_conf_context_keys = [
+            ('AWS_S3_BASE_URL',),
+            ('MEDIA_URL',),
+            ('MEDIA_PATH', os.path.join(remote_project_source_root, 'media')),
+            ('REDIRECT_STATIC_TO_S3',),
+            ('STATIC_URL',),
+            ('STATIC_PATH', os.path.join(remote_project_source_root, 'static')),
+            ('SERVE_FAVICON',),
+            ('SERVE_MEDIA',),
+            ('SERVE_ROBOTS',),
+            ('SERVE_STATIC',),
+            ('URL_PREFIX',),
+            ('VIRTUALENV_SITE_PACKAGES_PATH', default_site_packages),
+            ('WSGI_PROCESS_GROUP',),
+            ('WSGI_PROCESS_NAME', project_name),
+            ('WSGI_PROCESS_USER',),
+            ('WSGI_PROCESSES_COUNT',),
+            ('WSGI_PYTHON_OPTIMIZE_LEVEL',),
+            ('WSGI_THREADS_COUNT',),
+        ]
+        apache_conf_context = {}
+        for settings_fetch_args in apache_conf_context_keys:
+            context_value = deploy_settings.get(*settings_fetch_args)
+            apache_conf_context[settings_fetch_args[0].lower()] = context_value
+        apache_conf_context['project_path'] = remote_project_source_root
+        apache_conf_context['wsgi_script_dir_name'] = wsgi_script_dir_name
+        apache_conf_context['wsgi_script_file_name'] = wsgi_script_file_name
+        # Render Apache conf
+        apache_conf_content = render_file(apache_conf_template_path, apache_conf_context, optimize=True)
+        apache_conf_tmp_path = os.path.join(project_src_root, apache_conf_tmp_name)
+        with open(apache_conf_tmp_path, 'w') as f:
+            f.write(apache_conf_content)
+        puts('')
 
     # Send codebase
-    puts(cyan('Copy project codebase to remote'))
+    print_title('Copy project codebase to remote')
+    # Clean old pyc
+    target_run('find %s -name \'*.pyc\' -exec rm -rf {} \;'%remote_project_source_root)
+    # Do rsync
     rsync_argument = '-DLprtz --delete --delete-excluded --exclude \'*.pyc\' --exclude-from \'.gitignore\''
+    if verbose==3: rsync_argument += ' -v'
     local(rsync(
         project_src_root + ('' if project_src_root.endswith('/') else '/'),
-        remote_project_source_root,
+        (host_string+':' if is_remote else '') + remote_project_source_root,
         rsync_argument=rsync_argument
     ))
 
     # Copy gitignore
     with target_cd(remote_project_path):
-        target_run(rsync('source/.gitignore', '.gitignore'))
+        target_run(rsync('%s/.gitignore'%source_root_name, '.gitignore'))
 
     # Install venv
-    if getattr(deploy_settings, 'CHECK_PIP_REQUIREMENTS', True):
-        puts(cyan('Install virtualenv for your project'))
-        with fab_settings(virtual_env(remote_project_venv)):
-            target_run('pip install -r %s/requirements.txt --upgrade' % source_root_name)
+    if deploy_settings.get('CHECK_PIP_REQUIREMENTS'):
+        print_title('Install/Upgrade virtualenv for your project')
+        with fab_settings(virtual_env(remote_project_venv), target_cd(remote_project_path)):
+            target_run('pip install -r %s/requirements.txt'%source_root_name)
+
+    # Add timestamp in wsgi.py
+    wsgi_path = os.path.join(wsgi_script_dir_name, wsgi_script_file_name)
+    target_run('echo "# " >> %s'%wsgi_path)
+    target_run('echo "# Deploy at: %s" >> %s'%(datetime.datetime.now().strftime('%s.%f'), wsgi_path))
 
     # Compile python
-    if getattr(deploy_settings, 'COMPILE_PYC', True):
-        clean_py = getattr(deploy_settings, 'CLEAN_PY_IN_REMOTE', False)
-        puts(cyan('Compile to Python bytecode: ') + '%s'%remote_project_source_root)
+    if deploy_settings.get('COMPILE_PYC'):
+        clean_py = deploy_settings.get('CLEAN_PY_IN_REMOTE')
+        print_title('Compile Python bytecode')
         target_run('python -m compileall -q %s'%remote_project_source_root)
         if clean_py:
             target_run('find %s -name \'*.py\' -exec rm -rf {} \;'%remote_project_source_root)
 
-    # Update settings and commit git
+    # Install WSGI
+    apache_conf_updated = False
+    if deploy_settings.get('INSTALL_APACHE_CONF') and update_apache_conf:
+        print(cyan('Install apache conf'))
+        apache_conf_path = deploy_settings.get('APACHE_CONF_DIR')
+        apache_conf_target = os.path.join(apache_conf_path, '%s.conf'%project_name)
+        print cyan('Installing Apache conf')
+        with target_cd(remote_project_source_root):
+            target_sudo('mv %s %s'%(
+                apache_conf_tmp_name,
+                apache_conf_target
+            ))
+        apache_conf_updated = True
+
+    # Commit git
     puts(cyan('Update project record and commit repository'))
     with fab_settings(target_cd(remote_project_path), hide('warnings'), warn_only=True):
         target_run('git add -A && git commit -m "%s"'%commit_msg)
 
-    # Install apache conf
-    if getattr(deploy_settings, 'USE_MOD_WSGI', True):
-        puts(cyan('Install Apache configuration file'))
-        apache_conf_source = os.path.join(remote_project_source_root, env.tmp_apache_conf_path)
-        apache_conf_target = getattr(deploy_settings, 'APACHE_CONF_DIR', '/etc/httpd/conf.d/')
-        target_run('cp %s %s'%(apache_conf_source, apache_conf_target))
-
-    if getattr(deploy_settings, 'RESTART_APACHE', True):
-        puts(cyan('Restarting Apache HTTP Server'))
+    if deploy_settings.get('RESTART_APACHE') or apache_conf_updated:
+        print(cyan('Restarting Apache HTTP Server'))
         target_sudo('/etc/init.d/httpd restart')
 
     print ''
